RAW = 1
CCITTRLE = 2  # CCITT 10
CCITT_T4 = 3  # T4/Group 3 Fax
CCITT_T6 = 4  # T6/Group 4 Fax
LZW = 5  # Lempel-Ziv-Welch
JPEG_OLD = 6  # old-style JPEG
JPEG = 7
ADOBE_DEFLATE = 8
JBIG_BW = 9  # VC5
# JBIG_COLOR = 10
# JPEG_99 = 99
# KODAK_262 = 262
JPEGXR_NDPI = 22610  # JPEG XR (Hammamatsu NDPI)
# NEXT = 32766
# SONY_ARW = 32767
# PACKED_RAW = 32769
CCIRLEW = 32771  # Word-aligned 1D Huffman compression
PACKBITS = 32773  # PackBits, aka Macintosh RLE
THUNDERSCAN = 32809
# IT8LW = 32896
# IT8MP = 32897
# IT8BL = 32898
# PIXARFILM = 32908
# PIXARLOG = 32909
DEFLATE = 32946
DCS = 32947
JPEG2000_YUV = 33003  # JPEG 2000 YCbCr (Leica Aperio)
JPEG2000_LOSSY = 33004  # Lossy JPEG 2000 (Bio-Formats)
JPEG2000_RGB = 33005  # JPEG 2000 RGB (Leica Aperio)
ALT_JPEG = 33007  # JPEG (Bio-Formats)
JBIG = 34661
SGILOG = 34676  # LogLuv32
SGILOG24 = 34677
JPEG2000 = 34712
# NIKON_NEF = 32713
# JBIG2 = 32715
# MDI_BINARY = 34718  # Microsoft Document Imaging
# MDI_PROGRESSIVE = 34719  # Microsoft Document Imaging
# MDI_VECTOR = 34720  # Microsoft Document Imaging
LERC = 34887  # ESRI Limited Error Raster Compression
JPEG_LOSSY = 34892  # DNG
LZMA = 34925  # Lempel-Ziv-Markov chain Algorithm
ZSTD_DEPRECATED = 34926  # deprecated
WEBP_DEPRECATED = 34927  # deprecated
PNG = 34933  # Portable Network Graphics (Zoomable Image File format)
JPEGXR = 34934  # JPEG XR (Zoomable Image File format)
JETRAW = 48124  # Jetraw by Dotphoton
ZSTD = 50000  # Zstandard
WEBP = 50001  # WebP
JPEGXL = 50002  # GDAL
PIXTIFF = 50013  # ZLIB (Atalasoft)
JPEGXL_DNG = 52546  # JPEG XL (DNG)
EER_V0 = 65000  # FIXED82 Thermo Fisher Scientific
EER_V1 = 65001  # FIXED72 Thermo Fisher Scientific
EER_V2 = 65002  # VARIABLE Thermo Fisher Scientific
# PENTAX_PEF = 65535
